import React, { createContext, useContext, useRef, useEffect, useMemo, useReducer, useCallback } from "react";
import { createStubMethods, keysOf, createStubEffects } from "./utils";
import { tryCreateDevToolsLogger } from "./devtools";
export function useLocalux(initialState, methods, effectsIn = {}) {
    // To avoid warnings "Can't perform a React state update on an unmounted component"
    const isMounted = useRef(true);
    useEffect(() => () => {
        isMounted.current = false;
    }, []);
    const reducer = useMemo(() => {
        const logger = tryCreateDevToolsLogger();
        return (state, action) => {
            const nextState = methods[action.type](state)(...action.payload);
            if (logger) {
                logger(action, nextState);
            }
            return nextState;
        };
    }, []);
    const [state, dispatch] = useReducer(reducer, initialState);
    const stateRef = useRef(initialState);
    const getState = useCallback(() => stateRef.current, []);
    useEffect(() => {
        stateRef.current = state;
    }, [state]);
    const methodsOut = useMemo(() => keysOf(methods).reduce((methodsOutAcc, type) => {
        methodsOutAcc[type] = (...payload) => {
            // Do not dispatch actions on unmounted Provider
            if (!isMounted.current) {
                console.warn("Method called on unmounted Provider, but it can be ok.");
                return;
            }
            dispatch({
                type,
                payload
            });
        };
        return methodsOutAcc;
    }, {}), []);
    const effectsOut = useMemo(() => {
        return keysOf(effectsIn).reduce((effectsOutAcc, effectFnName) => {
            effectsOutAcc[effectFnName] = (...args) => {
                return effectsIn[effectFnName](getState, methodsOut)(...args);
            };
            return effectsOutAcc;
        }, {});
    }, [methodsOut]);
    return [state, methodsOut, effectsOut];
}
export const createUseStore = (defaultState, methodsIn, effectsIn) => {
    const methodsStubsWithoutProvider = createStubMethods(methodsIn);
    const effectsStubsWithoutProvider = createStubEffects(effectsIn);
    const context = createContext([
        defaultState,
        methodsStubsWithoutProvider,
        effectsStubsWithoutProvider
    ]);
    const StoreProvider = ({ initialState, children }) => {
        const [state, methods, effects] = useLocalux(initialState, methodsIn, effectsIn);
        return (React.createElement(context.Provider, { value: [state, methods, effects] }, children));
    };
    const useStore = () => {
        const [state, methods, effects] = useContext(context);
        return { state, methods, effects };
    };
    useStore.Provider = StoreProvider;
    return useStore;
};
